@@ -8,69 +8,77 @@
 // option. This file may not be copied, modified, or distributed
 // except according to those terms.
 
-/// Formatting of chained expressions, i.e. expressions which are chained by
-/// dots: struct and enum field access, method calls, and try shorthand (?).
-///
-/// Instead of walking these subexpressions one-by-one, as is our usual strategy
-/// for expression formatting, we collect maximal sequences of these expressions
-/// and handle them simultaneously.
-///
-/// Whenever possible, the entire chain is put on a single line. If that fails,
-/// we put each subexpression on a separate, much like the (default) function
-/// argument function argument strategy.
-///
-/// Depends on config options: `chain_indent` is the indent to use for
-/// blocks in the parent/root/base of the chain (and the rest of the chain's
-/// alignment).
-/// E.g., `let foo = { aaaa; bbb; ccc }.bar.baz();`, we would layout for the
-/// following values of `chain_indent`:
-/// Block:
-/// ```
-/// let foo = {
-///     aaaa;
-///     bbb;
-///     ccc
-/// }.bar
-///     .baz();
-/// ```
-/// Visual:
-/// ```
-/// let foo = {
-///               aaaa;
-///               bbb;
-///               ccc
-///           }
-///           .bar
-///           .baz();
-/// ```
-///
-/// If the first item in the chain is a block expression, we align the dots with
-/// the braces.
-/// Block:
-/// ```
-/// let a = foo.bar
-///     .baz()
-///     .qux
-/// ```
-/// Visual:
-/// ```
-/// let a = foo.bar
-///            .baz()
-///            .qux
-/// ```
+//! Formatting of chained expressions, i.e. expressions which are chained by
+//! dots: struct and enum field access, method calls, and try shorthand (?).
+//!
+//! Instead of walking these subexpressions one-by-one, as is our usual strategy
+//! for expression formatting, we collect maximal sequences of these expressions
+//! and handle them simultaneously.
+//!
+//! Whenever possible, the entire chain is put on a single line. If that fails,
+//! we put each subexpression on a separate, much like the (default) function
+//! argument function argument strategy.
+//!
+//! Depends on config options: `chain_indent` is the indent to use for
+//! blocks in the parent/root/base of the chain (and the rest of the chain's
+//! alignment).
+//! E.g., `let foo = { aaaa; bbb; ccc }.bar.baz();`, we would layout for the
+//! following values of `chain_indent`:
+//! Block:
+//!
+//! ```ignore
+//! let foo = {
+//!     aaaa;
+//!     bbb;
+//!     ccc
+//! }.bar
+//!     .baz();
+//! ```
+//!
+//! Visual:
+//!
+//! ```ignore
+//! let foo = {
+//!               aaaa;
+//!               bbb;
+//!               ccc
+//!           }
+//!           .bar
+//!           .baz();
+//! ```
+//!
+//! If the first item in the chain is a block expression, we align the dots with
+//! the braces.
+//! Block:
+//!
+//! ```ignore
+//! let a = foo.bar
+//!     .baz()
+//!     .qux
+//! ```
+//!
+//! Visual:
+//!
+//! ```ignore
+//! let a = foo.bar
+//!            .baz()
+//!            .qux
+//! ```
 
-use shape::Shape;
 use config::IndentStyle;
 use expr::rewrite_call;
 use macros::convert_try_mac;
 use rewrite::{Rewrite, RewriteContext};
+use shape::Shape;
 use utils::{first_line_width, last_line_extendable, last_line_width, mk_sp,
             trimmed_last_line_width, wrap_str};
 
+use std::borrow::Cow;
 use std::cmp::min;
 use std::iter;
-use syntax::{ast, ptr};
+
 use syntax::codemap::Span;
+use syntax::{ast, ptr};
 
 pub fn rewrite_chain(expr: &ast::Expr, context: &RewriteContext, shape: Shape) -> Option<String> {
     debug!("rewrite_chain {:?}", shape);
@@ -96,7 +104,7 @@ pub fn rewrite_chain(expr: &ast::Expr, context: &RewriteContext, shape: Shape) -
     };
     let parent_rewrite = parent
         .rewrite(context, parent_shape)
-        .map(|parent_rw| parent_rw + &repeat_try(prefix_try_num))?;
+        .map(|parent_rw| parent_rw + &"?".repeat(prefix_try_num))?;
     let parent_rewrite_contains_newline = parent_rewrite.contains('\n');
     let is_small_parent = parent_rewrite.len() <= context.config.tab_spaces();
 
@@ -129,15 +137,14 @@ pub fn rewrite_chain(expr: &ast::Expr, context: &RewriteContext, shape: Shape) -
         let offset = trimmed_last_line_width(&parent_rewrite) + prefix_try_num;
         match context.config.indent_style() {
             IndentStyle::Visual => parent_shape.offset_left(overhead)?,
-            IndentStyle::Block => parent_shape.block().offset_left(offset)?,
+            IndentStyle::Block => parent_shape.offset_left(offset)?,
         }
     } else {
         other_child_shape
     };
     debug!(
         "child_shapes {:?} {:?}",
-        first_child_shape,
-        other_child_shape
+        first_child_shape, other_child_shape
     );
 
     let child_shape_iter = Some(first_child_shape)
@@ -147,37 +154,33 @@ pub fn rewrite_chain(expr: &ast::Expr, context: &RewriteContext, shape: Shape) -
     let last_subexpr = &subexpr_list[suffix_try_num];
     let subexpr_list = &subexpr_list[suffix_try_num..subexpr_num - prefix_try_num];
     let iter = subexpr_list.iter().skip(1).rev().zip(child_shape_iter);
-    let mut rewrites = iter.map(|(e, shape)| {
-        rewrite_chain_subexpr(e, total_span, context, shape)
-    }).collect::<Option<Vec<_>>>()?;
+    let mut rewrites = iter.map(|(e, shape)| rewrite_chain_subexpr(e, total_span, context, shape))
+        .collect::<Option<Vec<_>>>()?;
 
     // Total of all items excluding the last.
-    let extend_last_subexpr = last_line_extendable(&parent_rewrite) && rewrites.is_empty();
+    let extend_last_subexpr = if is_small_parent {
+        rewrites.len() == 1 && last_line_extendable(&rewrites[0])
+    } else {
+        rewrites.is_empty() && last_line_extendable(&parent_rewrite)
+    };
     let almost_total = if extend_last_subexpr {
         last_line_width(&parent_rewrite)
     } else {
         rewrites.iter().fold(0, |a, b| a + b.len()) + parent_rewrite.len()
     } + suffix_try_num;
-    let one_line_budget = if rewrites.is_empty() && !context.config.chain_split_single_child() {
+    let one_line_budget = if rewrites.is_empty() {
         shape.width
     } else {
-        min(shape.width, context.config.chain_width())
+        min(shape.width, context.config.width_heuristics().chain_width)
     };
     let all_in_one_line = !parent_rewrite_contains_newline
         && rewrites.iter().all(|s| !s.contains('\n'))
         && almost_total < one_line_budget;
-    let last_shape = {
-        let last_shape = if rewrites.len() == 0 {
-            first_child_shape
-        } else {
-            other_child_shape
-        };
-        match context.config.indent_style() {
-            IndentStyle::Visual => last_shape.sub_width(shape.rhs_overhead(context.config))?,
-            IndentStyle::Block => last_shape,
-        }
-    };
-    let last_shape = last_shape.sub_width(suffix_try_num)?;
+    let last_shape = if rewrites.is_empty() {
+        first_child_shape
+    } else {
+        other_child_shape
+    }.sub_width(shape.rhs_overhead(context.config) + suffix_try_num)?;
 
     // Rewrite the last child. The last child of a chain requires special treatment. We need to
     // know whether 'overflowing' the last child make a better formatting:
@@ -251,13 +254,13 @@ pub fn rewrite_chain(expr: &ast::Expr, context: &RewriteContext, shape: Shape) -
 
     let connector = if fits_single_line && !parent_rewrite_contains_newline {
         // Yay, we can put everything on one line.
-        String::new()
+        Cow::from("")
     } else {
         // Use new lines.
-        if context.force_one_line_chain {
+        if *context.force_one_line_chain.borrow() {
             return None;
         }
-        format!("\n{}", nested_shape.indent.to_string(context.config))
+        nested_shape.indent.to_string_with_newline(context.config)
     };
 
     let first_connector = if is_small_parent || fits_single_line
@@ -266,7 +269,7 @@ pub fn rewrite_chain(expr: &ast::Expr, context: &RewriteContext, shape: Shape) -
     {
         ""
     } else {
-        connector.as_str()
+        &connector
     };
 
     let result = if is_small_parent && rewrites.len() > 1 {
@@ -294,7 +297,7 @@ pub fn rewrite_chain(expr: &ast::Expr, context: &RewriteContext, shape: Shape) -
             join_rewrites(&rewrites, &connector)
         )
     };
-    let result = format!("{}{}", result, repeat_try(suffix_try_num));
+    let result = format!("{}{}", result, "?".repeat(suffix_try_num));
     if context.config.indent_style() == IndentStyle::Visual {
         wrap_str(result, context.config.max_width(), shape)
     } else {
@@ -313,12 +316,6 @@ fn chain_only_try(exprs: &[ast::Expr]) -> bool {
     })
 }
 
-// Try to rewrite and replace the last non-try child. Return `true` if
-// replacing succeeds.
-fn repeat_try(try_count: usize) -> String {
-    iter::repeat("?").take(try_count).collect::<String>()
-}
-
 fn rewrite_try(
     expr: &ast::Expr,
     try_count: usize,
@@ -326,7 +323,7 @@ fn rewrite_try(
     shape: Shape,
 ) -> Option<String> {
     let sub_expr = expr.rewrite(context, shape.sub_width(try_count)?)?;
-    Some(format!("{}{}", sub_expr, repeat_try(try_count)))
+    Some(format!("{}{}", sub_expr, "?".repeat(try_count)))
 }
 
 fn join_rewrites(rewrites: &[String], connector: &str) -> String {
@@ -337,7 +334,7 @@ fn join_rewrites(rewrites: &[String], connector: &str) -> String {
         if rewrite != "?" {
             result.push_str(connector);
         }
-        result.push_str(&rewrite[..]);
+        result.push_str(&rewrite);
     }
 
     result
@@ -424,10 +421,12 @@ fn rewrite_chain_subexpr(
     context: &RewriteContext,
     shape: Shape,
 ) -> Option<String> {
-    let rewrite_element = |expr_str: String| if expr_str.len() <= shape.width {
-        Some(expr_str)
-    } else {
-        None
+    let rewrite_element = |expr_str: String| {
+        if expr_str.len() <= shape.width {
+            Some(expr_str)
+        } else {
+            None
+        }
     };
 
     match expr.node {
@@ -439,9 +438,9 @@ fn rewrite_chain_subexpr(
                 },
                 _ => &[],
             };
-            rewrite_method_call(segment.identifier, types, expressions, span, context, shape)
+            rewrite_method_call(segment.ident, types, expressions, span, context, shape)
         }
-        ast::ExprKind::Field(_, ref field) => rewrite_element(format!(".{}", field.node)),
+        ast::ExprKind::Field(_, ref field) => rewrite_element(format!(".{}", field.name)),
         ast::ExprKind::TupField(ref expr, ref field) => {
             let space = match expr.node {
                 ast::ExprKind::TupField(..) => " ",
